import numpy as np
import serial as S
import matplotlib.pyplot as plt
import cv2

class PySkin:
    """
    Class for e-skin Humanoids Homeworks
    Features:
        * Communication with electronics
        * Visualization functions
    To make your homework we suggest using this class however it is not enforced.
    Some methods are to be implemented by students as a homework.
    """

    def __init__(self, port, shape = (4,4), buffer_length = 1, fs=20, timeout=1):
        """
        Initalize serial and creates e-skin data buffers
        :param fs: Sample rate of measurement
        :param port: string at which e-skin electronics are connected
        :param shape: tuple e-skin taxel array shape
        :param buffer_length: float size of a circular buffer for signal processing in seconds,
                              size of a buffer is given as fs*buffer_buffer_length ROUNDED DOWN
        :param timeout: float timeout in seconds for serial communication
        """
        self.port = port
        self.shape = shape
        self.fs = fs
        self.num_of_taxels = shape[0]*shape[1]
        self.serial = S.Serial(self.port,timeout = timeout)
        self.data_buffer = np.zeros((int(self.fs*buffer_length),shape[0],shape[1]))      # data buffer ( R(t-N:t) ), indexing: [t,X,Y]
        self.data_0 = np.zeros((shape[0],shape[1]))                                      # initial reading ( R(t=0) )
        self.data = np.zeros((shape[0], shape[1]))                                       # one data frame ( R(t) )
        self.last_frame = self.data_buffer.shape[0]-1                                    # Circular buffer iterator, index of current R(t)

        self.fig = None
        self.ax = None

    def read_frame(self):
        """
        Reads serial port input and parses received data
        :return: read frame if reading was valid, None otherwise
        """
        reading_valid = True
        frame = np.zeros(self.shape)
        while(reading_valid):

            # print(f"Start {time.time()}")
            data_read = self.serial.readline() # Read until '\n'

            # print(f"End {time.time()}")
            if data_read == b'': # if timeout, return false
                print("READ ERROR: Timeout!")
                reading_valid = False

            else:
                data_read = data_read.decode()
                data_read = data_read.split(',')
                if data_read[0] != "S":
                    continue

                frame = np.array(data_read[1:-1]).astype(float).reshape(self.shape)
                break

        if reading_valid: # recieved right amount of data
            return frame
        else:
            reading_valid = False   # Perfectionism
            print("READ ERROR: invalid frame read!")
            return None

    def read_frames_and_average(self, cnt=1):
        """
        Reads set number of frames and computes an average
        :param cnt: number of frames to be averaged
        :return: Mean over read frames
        """
        avg = np.zeros(self.shape)
        for i in range(cnt):
            tmp = None
            while(tmp is None):
                tmp = self.read_frame()
            avg+=tmp/cnt
        return avg

    def read_and_update_buffer(self, cnt=1):
        """
        Reads and stores valid frame (averaged) into data buffer and updates last_frame iterator if reading was valid
        :return: True if reading was valid, False otherwise
        """
        frame = self.read_frames_and_average(cnt)
        if np.any(frame) == None:
            return False
        else:
            self.last_frame = (self.last_frame+1)%self.data_buffer.shape[0]    # increase the interator of circular buffer
            self.data_buffer[self.last_frame,:,:] = frame.copy()
            return True

    def visualize_frame(self,data, window_name = "e-skin data"):
        """
        Visualizes data as 2D array plot in non-blocking mode in 8-bit resolution
        :param data: frame to be visualized, data can be of type float (range 0-1), bool
        :param window_name: string name for openCV window
        :return frame8: image shown
        """
        maximum = None
        offset = 0
        k = -1
        if data.dtype == float: # If R/R_0 is used
            maximum = 1.
            offset = -1
            k = -1
        if data.dtype == bool: # Binary data are drawn
            maximum = True
            k = 1
        frame = np.zeros((self.shape[0] * 100, self.shape[1] * 100))
        for i in range(self.shape[0]):
            for j in range(self.shape[1]):
                tmp = 255 * (k*(float(data[i, j])+offset)/maximum)
                if tmp > 255:
                    tmp = 255
                elif tmp < 0:
                    tmp = 0
                frame[i * 100:(i + 1) * 100, j * 100:(j + 1) * 100] = tmp

        frame8 = np.uint8(frame)
        frame8 = cv2.applyColorMap(frame8, cv2.COLORMAP_INFERNO)
        cv2.imshow(window_name, frame8)
        return frame8

    def draw_time_series(self,data, taxel=[]):
        """
        Draws time series for given taxels, blocking
        :param taxel: list of tuple coordinates format: [(x1,y1),(x2,y2)]
        :param data: Time series for one taxel (3D array)
        :return:
        """
        self.fig, self.ax = plt.subplots(1, 1)
        self.ax.grid()
        self.ax.set_title(f"Taxel time series")
        T = np.linspace(0, 1 / self.fs * data.shape[0], data.shape[0])
        for t in taxel:
            self.ax.plot(T, data[:, t[0],t[1]], label=f"[{t[0]},{t[1]}]")
        self.ax.set_xlabel("t [s]")
        self.ax.set_ylabel("R/R_0 [-]")
        self.ax.legend()
        plt.show()

    def draw_all_time_series(self, data):
        """
        Draws all time series for all taxels, blocking
        :param data: Frame time series (3D array)
        :return:
        """
        coord = [(x,y) for x in range(4) for y in range(4)]
        self.draw_time_series(data, coord)
